import { OrderedSet, OrderedMap, Collection, Set } from "immutable";
import { TypeRef, TypeReconstituter } from "./TypeBuilder";
import { TypeNames } from "./TypeNames";
import { TypeAttributes } from "./TypeAttributes";
export declare type PrimitiveStringTypeKind = "string" | "date" | "time" | "date-time";
export declare type PrimitiveTypeKind = "none" | "any" | "null" | "bool" | "integer" | "double" | PrimitiveStringTypeKind;
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "object" | "map" | "intersection";
export declare function isPrimitiveStringTypeKind(kind: TypeKind): kind is PrimitiveStringTypeKind;
export declare function isNumberTypeKind(kind: TypeKind): kind is "integer" | "double";
export declare function isPrimitiveTypeKind(kind: TypeKind): kind is PrimitiveTypeKind;
export declare abstract class Type {
    readonly typeRef: TypeRef;
    readonly kind: TypeKind;
    constructor(typeRef: TypeRef, kind: TypeKind);
    readonly abstract children: OrderedSet<Type>;
    directlyReachableTypes<T>(setForType: (t: Type) => OrderedSet<T> | null): OrderedSet<T>;
    getAttributes(): TypeAttributes;
    readonly hasNames: boolean;
    getNames(): TypeNames;
    getCombinedName(): string;
    readonly abstract isNullable: boolean;
    abstract isPrimitive(): this is PrimitiveType;
    abstract map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    equals(other: any): boolean;
    hashCode(): number;
    protected abstract structuralEqualityStep(other: Type, queue: (a: Type, b: Type) => boolean): boolean;
    structurallyCompatible(other: Type): boolean;
    getParentTypes(): Set<Type>;
    getAncestorsNotInSet(set: Set<TypeRef>): Set<Type>;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(typeRef: TypeRef, kind: PrimitiveTypeKind, checkKind?: boolean);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class StringType extends PrimitiveType {
    readonly enumCases: OrderedMap<string, number> | undefined;
    constructor(typeRef: TypeRef, enumCases: OrderedMap<string, number> | undefined);
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ArrayType extends Type {
    private _itemsRef;
    readonly kind: "array";
    constructor(typeRef: TypeRef, _itemsRef?: TypeRef | undefined);
    setItems(itemsRef: TypeRef): undefined;
    private getItemsRef();
    readonly items: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: ArrayType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class GenericClassProperty<T> {
    readonly typeData: T;
    readonly isOptional: boolean;
    constructor(typeData: T, isOptional: boolean);
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class ClassProperty extends GenericClassProperty<TypeRef> {
    constructor(typeRef: TypeRef, isOptional: boolean);
    readonly typeRef: TypeRef;
    readonly type: Type;
}
export declare class ObjectType extends Type {
    readonly isFixed: boolean;
    readonly properties: OrderedMap<string, ClassProperty>;
    private _additionalPropertiesRef;
    constructor(typeRef: TypeRef, kind: TypeKind, isFixed: boolean, properties: OrderedMap<string, ClassProperty>, _additionalPropertiesRef: TypeRef | undefined);
    readonly sortedProperties: OrderedMap<string, ClassProperty>;
    readonly additionalProperties: Type | undefined;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: ObjectType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ClassType extends ObjectType {
    readonly isFixed: boolean;
    readonly properties: OrderedMap<string, ClassProperty>;
    kind: "class";
    constructor(typeRef: TypeRef, isFixed: boolean, properties: OrderedMap<string, ClassProperty>);
}
export declare class MapType extends ObjectType {
    readonly kind: "map";
    constructor(typeRef: TypeRef, valuesRef: TypeRef);
    readonly values: Type;
}
export declare function assertIsObject(t: Type): ObjectType;
export declare function assertIsClass(t: Type): ClassType;
export declare class EnumType extends Type {
    readonly cases: OrderedSet<string>;
    kind: "enum";
    constructor(typeRef: TypeRef, cases: OrderedSet<string>);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    map(builder: TypeReconstituter, _: (tref: TypeRef) => TypeRef): TypeRef;
    protected structuralEqualityStep(other: EnumType, _queue: (a: Type, b: Type) => void): boolean;
}
export declare abstract class SetOperationType extends Type {
    private _memberRefs;
    constructor(typeRef: TypeRef, kind: TypeKind, _memberRefs?: OrderedSet<TypeRef> | undefined);
    setMembers(memberRefs: OrderedSet<TypeRef>): void;
    protected getMemberRefs(): OrderedSet<TypeRef>;
    readonly members: OrderedSet<Type>;
    readonly sortedMembers: OrderedSet<Type>;
    readonly children: OrderedSet<Type>;
    isPrimitive(): this is PrimitiveType;
    protected structuralEqualityStep(other: UnionType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class IntersectionType extends SetOperationType {
    kind: "intersection";
    constructor(typeRef: TypeRef, memberRefs?: OrderedSet<TypeRef>);
    readonly isNullable: boolean;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare class UnionType extends SetOperationType {
    kind: "union";
    constructor(typeRef: TypeRef, memberRefs?: OrderedSet<TypeRef>);
    setMembers(memberRefs: OrderedSet<TypeRef>): void;
    readonly stringTypeMembers: OrderedSet<Type>;
    findMember(kind: TypeKind): Type | undefined;
    readonly isNullable: boolean;
    readonly isCanonical: boolean;
    map(builder: TypeReconstituter, f: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare function setOperationMembersRecursively<T extends SetOperationType>(setOperation: T): [OrderedSet<Type>, TypeAttributes];
export declare function setOperationMembersRecursively<T extends SetOperationType>(setOperations: T[]): [OrderedSet<Type>, TypeAttributes];
export declare function makeGroupsToFlatten<T extends SetOperationType>(setOperations: Set<T>, include: ((members: Set<Type>) => boolean) | undefined): Type[][];
export declare function combineTypeAttributesOfTypes(types: Collection<any, Type>): TypeAttributes;
export declare function setOperationCasesEqual(ma: OrderedSet<Type>, mb: OrderedSet<Type>, membersEqual: (a: Type, b: Type) => boolean): boolean;
export declare function removeNullFromUnion(t: UnionType, sortBy?: boolean | ((t: Type) => any)): [PrimitiveType | null, OrderedSet<Type>];
export declare function removeNullFromType(t: Type): [PrimitiveType | null, OrderedSet<Type>];
export declare function nullableFromUnion(t: UnionType): Type | null;
export declare function nonNullTypeCases(t: Type): OrderedSet<Type>;
export declare function getNullAsOptional(cp: ClassProperty): [boolean, OrderedSet<Type>];
export declare function isNamedType(t: Type): boolean;
export declare type SeparatedNamedTypes = {
    classes: OrderedSet<ClassType>;
    enums: OrderedSet<EnumType>;
    unions: OrderedSet<UnionType>;
};
export declare function separateNamedTypes(types: Collection<any, Type>): SeparatedNamedTypes;
export declare function directlyReachableSingleNamedType(type: Type): Type | undefined;
export declare type StringTypeMatchers<U> = {
    dateType?: (dateType: PrimitiveType) => U;
    timeType?: (timeType: PrimitiveType) => U;
    dateTimeType?: (dateTimeType: PrimitiveType) => U;
};
export declare function matchTypeExhaustive<U>(t: Type, noneType: (noneType: PrimitiveType) => U, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: StringType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, objectType: (objectType: ObjectType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, dateType: (dateType: PrimitiveType) => U, timeType: (timeType: PrimitiveType) => U, dateTimeType: (dateTimeType: PrimitiveType) => U): U;
export declare function matchType<U>(type: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: StringType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, stringTypeMatchers?: StringTypeMatchers<U>): U;
export declare function matchCompoundType(t: Type, arrayType: (arrayType: ArrayType) => void, classType: (classType: ClassType) => void, mapType: (mapType: MapType) => void, unionType: (unionType: UnionType) => void): void;
